@@ -29,19 +29,23 @@
 #include <glib/gi18n.h>
 #include <gdk/gdk.h>
 #include <gst/tag/tag.h>
-#ifdef HAVE_GSTREAMER_0_8
-#include <gst/gconf/gconf.h>
-#endif
 #include <libgnomevfs/gnome-vfs-ops.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
 
-#include "rb-debug.h"
-#include "rb-player.h"
-#include "rb-player-gst.h"
 #include "rb-debug.h"
 #include "rb-marshal.h"
+#include "rb-util.h"
+
+#include "rb-player.h"
+#include "rb-player-gst-filter.h"
+#include "rb-player-gst-tee.h"
+/*#include "rb-player-gst-data-tee.h"*/
+#include "rb-player-gst.h"
 
 static void rb_player_init (RBPlayerIface *iface);
+static void rb_player_gst_filter_init (RBPlayerGstFilterIface *iface);
+static void rb_player_gst_tee_init (RBPlayerGstTeeIface *iface);
+/*tatic void rb_player_gst_data_tee_init (RBPlayerGstDataTeeIface *iface);*/
 static void rb_player_gst_finalize (GObject *object);
 static void rb_player_gst_get_property (GObject *object,
 					guint prop_id,
@@ -50,24 +54,32 @@ static void rb_player_gst_get_property (GObject *object,
 
 static gboolean rb_player_gst_open (RBPlayer *player,
 				    const char *uri,
+				    gpointer stream_data,
+				    GDestroyNotify stream_data_destroy,
 				    GError **error);
 static gboolean rb_player_gst_opened (RBPlayer *player);
-static gboolean rb_player_gst_close (RBPlayer *player, GError **error);
-static gboolean rb_player_gst_play (RBPlayer *player, GError **error);
+static gboolean rb_player_gst_close (RBPlayer *player,
+				     const char *uri,
+				     GError **error);
+static gboolean rb_player_gst_play (RBPlayer *player, gint crossfade, GError **error);
 static void rb_player_gst_pause (RBPlayer *player);
 static gboolean rb_player_gst_playing (RBPlayer *player);
 static void rb_player_gst_set_volume (RBPlayer *player, float volume);
 static float rb_player_gst_get_volume (RBPlayer *player);
 static void rb_player_gst_set_replaygain (RBPlayer *player,
-				      double track_gain, double track_peak,
-				      double album_gain, double album_peak);
+					  const char *uri,
+					  double track_gain, double track_peak,
+					  double album_gain, double album_peak);
 static gboolean rb_player_gst_seekable (RBPlayer *player);
 static void rb_player_gst_set_time (RBPlayer *player, long time);
 static long rb_player_gst_get_time (RBPlayer *player);
 
 G_DEFINE_TYPE_WITH_CODE(RBPlayerGst, rb_player_gst, G_TYPE_OBJECT,
-			G_IMPLEMENT_INTERFACE(RB_TYPE_PLAYER,
-					      rb_player_init))
+			G_IMPLEMENT_INTERFACE(RB_TYPE_PLAYER, rb_player_init)
+			G_IMPLEMENT_INTERFACE(RB_TYPE_PLAYER_GST_FILTER, rb_player_gst_filter_init)
+			G_IMPLEMENT_INTERFACE(RB_TYPE_PLAYER_GST_TEE, rb_player_gst_tee_init)
+			/*G_IMPLEMENT_INTERFACE(RB_TYPE_PLAYER_GST_DATA_TEE, rb_player_gst_data_tee_init)*/
+			)
 #define RB_PLAYER_GST_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), RB_TYPE_PLAYER_GST, RBPlayerGstPrivate))
 
 #define RB_PLAYER_GST_TICK_HZ 5
@@ -81,6 +93,8 @@ enum
 struct _RBPlayerGstPrivate
 {
 	char *uri;
+	gpointer stream_data;
+	GDestroyNotify stream_data_destroy;
 
 	GstElement *playbin;
 
@@ -96,10 +110,12 @@ struct _RBPlayerGstPrivate
 	guint idle_buffering_id;
 	GHashTable *idle_info_ids;
 
-#ifdef HAVE_GSTREAMER_0_8
-	guint error_signal_id;
-	guint buffering_signal_id;
-#endif
+	GList *waiting_tees;
+	GstElement *sinkbin;
+	GstElement *tee;
+
+	GList *waiting_filters; /* in reverse order */
+	GstElement *filterbin;
 
 	float cur_volume;
 
@@ -163,6 +179,7 @@ rb_player_init (RBPlayerIface *iface)
 	iface->seekable = rb_player_gst_seekable;
 	iface->set_time = rb_player_gst_set_time;
 	iface->get_time = rb_player_gst_get_time;
+	iface->multiple_open = (RBPlayerFeatureFunc) rb_false_function;
 }
 
 static gboolean
@@ -171,7 +188,7 @@ tick_timeout (RBPlayerGst *mp)
 	if (mp->priv->playing == FALSE)
 		return TRUE;
 
-	_rb_player_emit_tick (RB_PLAYER (mp), rb_player_get_time (RB_PLAYER (mp)));
+	_rb_player_emit_tick (RB_PLAYER (mp), mp->priv->stream_data, rb_player_get_time (RB_PLAYER (mp)), -1);
 
 	return TRUE;
 }
@@ -179,11 +196,7 @@ tick_timeout (RBPlayerGst *mp)
 static void
 rb_player_gst_init (RBPlayerGst *mp)
 {
-	gint ms_period = 1000 / RB_PLAYER_GST_TICK_HZ;
-
 	mp->priv = RB_PLAYER_GST_GET_PRIVATE (mp);
-
-	mp->priv->tick_timeout_id = g_timeout_add (ms_period, (GSourceFunc) tick_timeout, mp);
 	mp->priv->idle_info_ids = g_hash_table_new (NULL, NULL);
 
 }
@@ -213,23 +226,22 @@ rb_player_gst_finalize (GObject *object)
 
 	mp = RB_PLAYER_GST (object);
 
-	g_source_remove (mp->priv->tick_timeout_id);
+	if (mp->priv->tick_timeout_id != 0)
+		g_source_remove (mp->priv->tick_timeout_id);
 	g_hash_table_destroy (mp->priv->idle_info_ids);
 
 	if (mp->priv->playbin) {
-#ifdef HAVE_GSTREAMER_0_8
-		g_signal_handler_disconnect (G_OBJECT (mp->priv->playbin),
-					     mp->priv->error_signal_id);
-		g_signal_handler_disconnect (G_OBJECT (mp->priv->playbin),
-					     mp->priv->buffering_signal_id);
-#endif
-
 		gst_element_set_state (mp->priv->playbin,
 				       GST_STATE_NULL);
 
 		rb_player_gst_gst_free_playbin (mp);
 	}
 
+	if (mp->priv->waiting_tees) {
+		g_list_foreach (mp->priv->waiting_tees, (GFunc)gst_object_sink, NULL);
+	}
+	g_list_free (mp->priv->waiting_tees);
+
 	G_OBJECT_CLASS (rb_player_gst_parent_class)->finalize (object);
 }
 
@@ -271,117 +283,46 @@ emit_signal_idle (RBPlayerGstSignal *signal)
 {
 	switch (signal->type) {
 	case ERROR:
-		_rb_player_emit_error (RB_PLAYER (signal->object), signal->error);
+		_rb_player_emit_error (RB_PLAYER (signal->object),
+				       signal->object->priv->stream_data,
+				       signal->error);
 
 		/* close if not already closing */
 		if (signal->object->priv->uri != NULL)
-			rb_player_close (RB_PLAYER (signal->object), NULL);
+			rb_player_close (RB_PLAYER (signal->object), NULL, NULL);
 
 		break;
 
 	case EOS:
-		_rb_player_emit_eos (RB_PLAYER (signal->object));
+		_rb_player_emit_eos (RB_PLAYER (signal->object),
+				     signal->object->priv->stream_data);
 		signal->object->priv->idle_eos_id = 0;
 		break;
 
 	case INFO:
-		_rb_player_emit_info (RB_PLAYER (signal->object), signal->info_field, signal->info);
+		_rb_player_emit_info (RB_PLAYER (signal->object),
+				      signal->object->priv->stream_data,
+				      signal->info_field,
+				      signal->info);
 		break;
 
 	case BUFFERING:
-		_rb_player_emit_buffering (RB_PLAYER (signal->object), g_value_get_uint (signal->info));
+		_rb_player_emit_buffering (RB_PLAYER (signal->object),
+					   signal->object->priv->stream_data,
+					   g_value_get_uint (signal->info));
 		signal->object->priv->idle_buffering_id = 0;
 		break;
 	case EVENT:
-		_rb_player_emit_event (RB_PLAYER (signal->object), g_value_get_string (signal->info), NULL);
+		_rb_player_emit_event (RB_PLAYER (signal->object),
+				       signal->object->priv->stream_data,
+				       g_value_get_string (signal->info),
+				       NULL);
 		break;
 	}
 
 	return FALSE;
 }
 
-#ifdef HAVE_GSTREAMER_0_8
-static void
-eos_cb (GstElement *element,
-	RBPlayerGst *mp)
-{
-	RBPlayerGstSignal *signal;
-	signal = g_new0 (RBPlayerGstSignal, 1);
-	signal->type = EOS;
-	signal->object = mp;
-	g_object_ref (G_OBJECT (mp));
-	if (mp->priv->idle_eos_id)
-		g_source_remove (mp->priv->idle_eos_id);
-	mp->priv->idle_eos_id = g_idle_add_full (G_PRIORITY_HIGH_IDLE,
-						 (GSourceFunc) emit_signal_idle,
-						 signal,
-						 destroy_idle_signal);
-}
-
-static void
-error_cb (GstElement *element,
-	  GstElement *source,
-	  const GError *error,
-	  gchar *debug,
-	  RBPlayerGst *mp)
-{
-	int code;
-	RBPlayerGstSignal *signal;
-
-	/* the handler shouldn't get called with error=NULL, but sometimes it does */
-	if (error == NULL)
-		return;
-
-	if ((error->domain == GST_CORE_ERROR)
-	    || (error->domain == GST_LIBRARY_ERROR)
-	    || (error->code == GST_RESOURCE_ERROR_BUSY)) {
-		code = RB_PLAYER_ERROR_NO_AUDIO;
-	} else {
-		code = RB_PLAYER_ERROR_GENERAL;
-	}
-
-	/* If we've already got an error, ignore 'internal data flow error'
-	 * type messages, as they're too generic to be helpful.
-	 */
-	if (mp->priv->emitted_error &&
-	    error->domain == GST_STREAM_ERROR &&
-	    error->code == GST_STREAM_ERROR_FAILED) {
-		rb_debug ("Ignoring generic error \"%s\"", error->message);
-		return;
-	}
-
-	/* If we're in a synchronous op, we can signal the error directly */
-	if (mp->priv->can_signal_direct_error) {
-		if (mp->priv->error) {
-			g_warning ("Overwriting previous error \"%s\" with new error \"%s\"",
-				   mp->priv->error->message,
-				   error->message);
-			g_error_free (mp->priv->error);
-		}
-		mp->priv->emitted_error = TRUE;
-		mp->priv->error = g_error_copy (error);
-		return;
-	}
-
-	signal = g_new0 (RBPlayerGstSignal, 1);
-	signal->type = ERROR;
-	signal->object = mp;
-	signal->error = g_error_new_literal (RB_PLAYER_ERROR,
-					     code,
-					     error->message);
-
-	g_object_ref (G_OBJECT (mp));
-
-	if (mp->priv->idle_error_id)
-		g_source_remove (mp->priv->idle_error_id);
-	mp->priv->idle_error_id = g_idle_add_full (G_PRIORITY_HIGH_IDLE,
-						   (GSourceFunc) emit_signal_idle,
-						   signal,
-						   destroy_idle_signal);
-	mp->priv->emitted_error = TRUE;
-}
-#endif
-
 static void
 process_tag (const GstTagList *list, const gchar *tag, RBPlayerGst *player)
 {
@@ -448,32 +389,6 @@ process_tag (const GstTagList *list, const gchar *tag, RBPlayerGst *player)
 	g_hash_table_insert (player->priv->idle_info_ids, GUINT_TO_POINTER (signal->id), NULL);
 }
 
-#ifdef HAVE_GSTREAMER_0_8
-static void
-found_tag_cb (GObject *pipeline, GstElement *source, GstTagList *tags, RBPlayerGst *player)
-{
-	gst_tag_list_foreach (tags, (GstTagForeachFunc) process_tag, player);
-}
-
-static void
-buffering_cb (GstElement *element, gint progress, RBPlayerGst *mp)
-{
-	RBPlayerGstSignal *signal;
-
-	signal = g_new0 (RBPlayerGstSignal, 1);
-	signal->type = BUFFERING;
-
-	g_object_ref (G_OBJECT (mp));
-	signal->object = mp;
-
-	signal->info = g_new0 (GValue, 1);
-	g_value_init (signal->info, G_TYPE_UINT);
-	g_value_set_uint (signal->info, (guint)progress);
-	if (mp->priv->idle_buffering_id)
-		g_source_remove (mp->priv->idle_buffering_id);
-	mp->priv->idle_buffering_id = g_idle_add ((GSourceFunc) emit_signal_idle, signal);
-}
-#elif HAVE_GSTREAMER_0_10
 static gboolean
 rb_player_gst_bus_cb (GstBus * bus, GstMessage * message, RBPlayerGst *mp)
 {
@@ -511,19 +426,19 @@ rb_player_gst_bus_cb (GstBus * bus, GstMessage * message, RBPlayerGst *mp)
 							 code,
 							 error->message);
 			mp->priv->emitted_error = TRUE;
-			_rb_player_emit_error (RB_PLAYER (mp), sig_error);
+			_rb_player_emit_error (RB_PLAYER (mp), mp->priv->stream_data, sig_error);
 		}
 
 		/* close if not already closing */
 		if (mp->priv->uri != NULL)
-			rb_player_close (RB_PLAYER (mp), NULL);
+			rb_player_close (RB_PLAYER (mp), NULL, NULL);
 
 		g_error_free (error);
 		g_free (debug);
 		break;
 	}
 	case GST_MESSAGE_EOS:
-		_rb_player_emit_eos (RB_PLAYER (mp));
+		_rb_player_emit_eos (RB_PLAYER (mp), mp->priv->stream_data);
 		break;
 	case GST_MESSAGE_TAG: {
 		GstTagList *tags;
@@ -596,8 +511,6 @@ rb_player_gst_bus_cb (GstBus * bus, GstMessage * message, RBPlayerGst *mp)
 	return TRUE;
 }
 
-#endif
-
 static gboolean
 rb_player_gst_construct (RBPlayerGst *mp, GError **error)
 {
@@ -612,34 +525,6 @@ rb_player_gst_construct (RBPlayerGst *mp, GError **error)
 		goto missing_element;
 	}
 
-#ifdef HAVE_GSTREAMER_0_8
-	g_signal_connect_object (G_OBJECT (mp->priv->playbin),
-				 "found_tag",
-				 G_CALLBACK (found_tag_cb),
-				 mp, 0);
-
-	mp->priv->buffering_signal_id =
-		g_signal_connect_object (G_OBJECT (mp->priv->playbin),
-					 "buffering",
-					 G_CALLBACK (buffering_cb),
-					 mp, 0);
-
-	mp->priv->error_signal_id =
-		g_signal_connect_object (G_OBJECT (mp->priv->playbin),
-					 "error",
-					 G_CALLBACK (error_cb),
-					 mp, 0);
-
-	g_signal_connect_object (G_OBJECT (mp->priv->playbin), "eos",
-				 G_CALLBACK (eos_cb), mp, 0);
-#endif
-
-#ifdef HAVE_GSTREAMER_0_8
-	/* Output sink */
-	sink = gst_gconf_get_default_audio_sink ();
-	g_object_set (G_OBJECT (mp->priv->playbin), "audio-sink", sink, NULL);
-#endif
-#ifdef HAVE_GSTREAMER_0_10
 	gst_bus_add_watch (gst_element_get_bus (GST_ELEMENT (mp->priv->playbin)),
 			     (GstBusFunc) rb_player_gst_bus_cb, mp);
 
@@ -665,7 +550,61 @@ rb_player_gst_construct (RBPlayerGst *mp, GError **error)
 	} else {
 		g_object_unref (sink);
 	}
-#endif
+
+	{
+		GstPad *pad;
+		GList *l;
+		GstElement *queue;
+		GstElement *audioconvert;
+		GstPad *ghostpad;
+
+		/* setup filterbin,and insert the leading audioconvert */
+		mp->priv->filterbin = gst_bin_new (NULL);
+		audioconvert = gst_element_factory_make ("audioconvert", NULL);
+		gst_bin_add (GST_BIN (mp->priv->filterbin), audioconvert);
+
+		/* ghost it to the bin */
+		pad = gst_element_get_pad (audioconvert, "sink");
+		ghostpad = gst_ghost_pad_new ("sink", pad);
+		gst_element_add_pad (mp->priv->filterbin, ghostpad);
+		gst_object_unref (pad);
+
+		pad = gst_element_get_pad (audioconvert, "src");
+		ghostpad = gst_ghost_pad_new ("src", pad);
+		gst_element_add_pad (mp->priv->filterbin, ghostpad);
+		gst_object_unref (pad);
+
+
+		/* set up the sinkbin with it's tee element */
+		mp->priv->sinkbin = gst_bin_new (NULL);
+		mp->priv->tee = gst_element_factory_make ("tee", NULL);
+		queue = gst_element_factory_make ("queue", NULL);
+
+		/* link it all together and insert */
+		gst_bin_add_many (GST_BIN (mp->priv->sinkbin), mp->priv->filterbin, mp->priv->tee, queue, sink, NULL);
+		gst_element_link_many (mp->priv->filterbin, mp->priv->tee, queue, sink, NULL);
+
+		pad = gst_element_get_pad (mp->priv->filterbin, "sink");
+		ghostpad = gst_ghost_pad_new ("sink", pad);
+		gst_element_add_pad (mp->priv->sinkbin, ghostpad);
+		gst_object_unref (pad);
+
+		g_object_set (G_OBJECT (mp->priv->playbin), "audio-sink", mp->priv->sinkbin, NULL);
+
+
+		/* add any tees and filters that were waiting for us */
+		for (l = mp->priv->waiting_tees; l != NULL; l = g_list_next (l)) {
+			rb_player_gst_tee_add_tee (RB_PLAYER_GST_TEE (mp), GST_ELEMENT (l->data));
+		}
+		g_list_free (mp->priv->waiting_tees);
+		mp->priv->waiting_tees = NULL;
+
+		for (l = mp->priv->waiting_filters; l != NULL; l = g_list_next (l)) {
+			rb_player_gst_filter_add_filter (RB_PLAYER_GST_FILTER(mp), GST_ELEMENT (l->data));
+		}
+		g_list_free (mp->priv->waiting_filters);
+		mp->priv->waiting_filters = NULL;
+	}
 
 	/* Use fakesink for video if there's no video sink yet */
 	g_object_get (G_OBJECT (mp->priv->playbin), "video-sink", &sink, NULL);
@@ -714,28 +653,18 @@ rb_player_gst_sync_pipeline (RBPlayerGst *mp)
 	rb_debug ("syncing pipeline");
 	if (mp->priv->playing) {
  		rb_debug ("PLAYING pipeline");
-#ifdef HAVE_GSTREAMER_0_8
- 		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PLAYING) == GST_STATE_FAILURE) {
-#elif HAVE_GSTREAMER_0_10
 		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE) {
-#endif
  			return FALSE;
 		}
 	} else {
 		rb_debug ("PAUSING pipeline");
-#ifdef HAVE_GSTREAMER_0_8
- 		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED) == GST_STATE_FAILURE) {
-#elif HAVE_GSTREAMER_0_10
 		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED) == GST_STATE_CHANGE_FAILURE) {
-#endif
 			return FALSE;
 		}
 	}
-#ifdef HAVE_GSTREAMER_0_10
 	/* FIXME: Set up a timeout to watch if the pipeline doesn't
-         * go to PAUSED/PLAYING within some time (5 secs maybe?)
+	 * go to PAUSED/PLAYING within some time (5 secs maybe?)
 	 */
-#endif
 	return TRUE;
 }
 
@@ -772,11 +701,7 @@ cdda_got_source_cb (GObject *object, GParamSpec *pspec, char *device)
 {
 	GstElement *source;
 
-#ifdef HAVE_GSTREAMER_0_10
 	g_object_get (object, "source", &source, NULL);
-#elif HAVE_GSTREAMER_0_8
-	gst_element_get (GST_ELEMENT (object), "source", &source, NULL);
-#endif
 	rb_debug ("got source %p", source);
 	if (source) {
 		g_signal_handlers_disconnect_by_func (object, cdda_got_source_cb, device);
@@ -792,9 +717,21 @@ cdda_got_source_cb (GObject *object, GParamSpec *pspec, char *device)
 	}
 }
 
+static void
+_destroy_stream_data (RBPlayerGst *player)
+{
+	if (player->priv->stream_data && player->priv->stream_data_destroy) {
+		player->priv->stream_data_destroy (player->priv->stream_data);
+	}
+	player->priv->stream_data = NULL;
+	player->priv->stream_data_destroy = NULL;
+}
+
 static gboolean
 rb_player_gst_open (RBPlayer *player,
 		const char *uri,
+		gpointer stream_data,
+		GDestroyNotify stream_data_destroy,
 		GError **error)
 {
 	RBPlayerGst *mp = RB_PLAYER_GST (player);
@@ -804,13 +741,14 @@ rb_player_gst_open (RBPlayer *player,
 		if (!rb_player_gst_construct (mp, error))
 			return FALSE;
 	} else {
-		if (!rb_player_close (player, error))
+		if (!rb_player_close (player, NULL, error))
 			return FALSE;
 	}
 
 	g_assert (mp->priv->playbin != NULL);
 
 	if (uri == NULL) {
+		_destroy_stream_data (mp);
 		g_free (mp->priv->uri);
 		mp->priv->uri = NULL;
 		mp->priv->playing = FALSE;
@@ -841,22 +779,11 @@ rb_player_gst_open (RBPlayer *player,
 
 			rb_debug ("seeking to track %d on CD device %s", track, new_device);
 
-#ifdef HAVE_GSTREAMER_0_10
 			if (gst_element_seek (mp->priv->playbin, 1.0,
 					      track_format, GST_SEEK_FLAG_FLUSH,
 					      GST_SEEK_TYPE_SET, track,
 					      GST_SEEK_TYPE_NONE, -1))
 				cdda_seek = TRUE;
-#elif HAVE_GSTREAMER_0_8
-			{
-				GstEvent *event;
-
-				event = gst_event_new_seek (track_format | GST_SEEK_METHOD_SET | GST_SEEK_FLAG_FLUSH,
-						    (guint64) track);
-				if (gst_element_send_event (mp->priv->playbin, event))
-					cdda_seek = TRUE;
-			}
-#endif
 		} else {
 			/* +1 to skip the '#' */
 			char *device = g_strdup (new_device + 1);
@@ -870,8 +797,11 @@ rb_player_gst_open (RBPlayer *player,
 	}
 
 	begin_gstreamer_operation (mp);
+	_destroy_stream_data (mp);
 	g_free (mp->priv->uri);
 	mp->priv->uri = g_strdup (uri);
+	mp->priv->stream_data = stream_data;
+	mp->priv->stream_data_destroy = stream_data_destroy;
 	mp->priv->emitted_error = FALSE;
 
 	if (!cdda_seek) {
@@ -880,10 +810,13 @@ rb_player_gst_open (RBPlayer *player,
 
 	if (!rb_player_gst_sync_pipeline (mp)) {
 		end_gstreamer_operation (mp, TRUE, error);
-		rb_player_gst_close (player, NULL);
+		rb_player_gst_close (player, uri, NULL);
 		return FALSE;
 	}
 
+	if (mp->priv->tick_timeout_id == 0)
+		mp->priv->tick_timeout_id = g_timeout_add (1000 / RB_PLAYER_GST_TICK_HZ, (GSourceFunc) tick_timeout, mp);
+
 	end_gstreamer_operation (mp, FALSE, error);
 	return TRUE;
 }
@@ -897,7 +830,7 @@ remove_idle_source (gpointer key, gpointer value, gpointer user_data)
 }
 
 static gboolean
-rb_player_gst_close (RBPlayer *player, GError **error)
+rb_player_gst_close (RBPlayer *player, const char *uri, GError **error)
 {
 	RBPlayerGst *mp = RB_PLAYER_GST (player);
 	gboolean ret;
@@ -905,6 +838,12 @@ rb_player_gst_close (RBPlayer *player, GError **error)
 	mp->priv->playing = FALSE;
 	mp->priv->buffering = FALSE;
 
+	if ((uri != NULL) && (mp->priv->uri != NULL) && strcmp (mp->priv->uri, uri) == 0) {
+		rb_debug ("URI doesn't match current playing URI; ignoring");
+		return TRUE;
+	}
+
+	_destroy_stream_data (mp);
 	g_free (mp->priv->uri);
 	mp->priv->uri = NULL;
 
@@ -922,15 +861,16 @@ rb_player_gst_close (RBPlayer *player, GError **error)
 	}
 	g_hash_table_foreach (mp->priv->idle_info_ids, remove_idle_source, NULL);
 
+	if (mp->priv->tick_timeout_id != 0) {
+		g_source_remove (mp->priv->tick_timeout_id);
+		mp->priv->tick_timeout_id = 0;
+	}
+
 	if (mp->priv->playbin == NULL)
 		return TRUE;
 
 	begin_gstreamer_operation (mp);
-#ifdef HAVE_GSTREAMER_0_8
-	ret = gst_element_set_state (mp->priv->playbin, GST_STATE_READY) == GST_STATE_SUCCESS;
-#elif HAVE_GSTREAMER_0_10
 	ret = gst_element_set_state (mp->priv->playbin, GST_STATE_READY) == GST_STATE_CHANGE_SUCCESS;
-#endif
 	end_gstreamer_operation (mp, !ret, error);
 	return ret;
 }
@@ -944,7 +884,7 @@ rb_player_gst_opened (RBPlayer *player)
 }
 
 static gboolean
-rb_player_gst_play (RBPlayer *player, GError **error)
+rb_player_gst_play (RBPlayer *player, gint crossfade, GError **error)
 {
 	RBPlayerGst *mp = RB_PLAYER_GST (player);
 	gboolean ret;
@@ -957,6 +897,12 @@ rb_player_gst_play (RBPlayer *player, GError **error)
 	begin_gstreamer_operation (mp);
 	ret = rb_player_gst_sync_pipeline (mp);
 	end_gstreamer_operation (mp, !ret, error);
+
+	_rb_player_emit_playing_stream (RB_PLAYER (mp), mp->priv->stream_data);
+
+	if (mp->priv->tick_timeout_id == 0)
+		mp->priv->tick_timeout_id = g_timeout_add (1000 / RB_PLAYER_GST_TICK_HZ, (GSourceFunc) tick_timeout, mp);
+
 	return ret;
 }
 
@@ -973,6 +919,11 @@ rb_player_gst_pause (RBPlayer *player)
 	g_return_if_fail (mp->priv->playbin != NULL);
 
 	rb_player_gst_sync_pipeline (mp);
+
+	if (mp->priv->tick_timeout_id != 0) {
+		g_source_remove (mp->priv->tick_timeout_id);
+		mp->priv->tick_timeout_id = 0;
+	}
 }
 
 static gboolean
@@ -985,6 +936,7 @@ rb_player_gst_playing (RBPlayer *player)
 
 static void
 rb_player_gst_set_replaygain (RBPlayer *player,
+			      const char *uri,
 			      double track_gain, double track_peak,
 			      double album_gain, double album_peak)
 {
@@ -1016,7 +968,7 @@ rb_player_gst_set_replaygain (RBPlayer *player,
 	if (scale > 15)
 		scale = 15;
 
-        rb_debug ("Scale : %f New volume : %f", scale, mp->priv->cur_volume * scale);
+	rb_debug ("Scale : %f New volume : %f", scale, mp->priv->cur_volume * scale);
 
 	if (mp->priv->playbin != NULL) {
 		GParamSpec *volume_pspec;
@@ -1064,20 +1016,13 @@ static gboolean
 rb_player_gst_seekable (RBPlayer *player)
 {
 	RBPlayerGst *mp = RB_PLAYER_GST (player);
-#ifdef HAVE_GSTREAMER_0_8
-	/* FIXME we're lying here,
-	 * (0.8) trying a seek might disrupt playback */
-	gboolean can_seek = TRUE;
-#elif HAVE_GSTREAMER_0_10
-	/* (0.10) Need to send a seekable query on playbin */
+	/* Need to send a seekable query on playbin */
 	gboolean can_seek = TRUE;
 	GstQuery *query;
-#endif
 
 	if (mp->priv->playbin == NULL)
 		return FALSE;
 
-#ifdef HAVE_GSTREAMER_0_10
 	query = gst_query_new_seeking (GST_FORMAT_TIME);
 	if (gst_element_query (mp->priv->playbin, query)) {
 		gst_query_parse_seeking (query, NULL, &can_seek, NULL, NULL);
@@ -1088,7 +1033,6 @@ rb_player_gst_seekable (RBPlayer *player)
 		can_seek = gst_element_query (mp->priv->playbin, query);
 	}
 	gst_query_unref (query);
-#endif
 
 	return can_seek;
 }
@@ -1101,29 +1045,17 @@ rb_player_gst_set_time (RBPlayer *player, long time)
 
 	g_return_if_fail (mp->priv->playbin != NULL);
 
-#ifdef HAVE_GSTREAMER_0_8
-	gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED);
-#elif HAVE_GSTREAMER_0_10
 	if (gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED) == GST_STATE_CHANGE_ASYNC) {
 		/* FIXME: Use a timeout on get_state. Post a GError somewhere on failed? */
 		if (gst_element_get_state (mp->priv->playbin, NULL, NULL, 3 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
 			g_warning ("Failed to pause pipeline before seek");
 		}
 	}
-#endif
 
-#ifdef HAVE_GSTREAMER_0_8
-	gst_element_seek (mp->priv->playbin,
-			  GST_FORMAT_TIME
-			  | GST_SEEK_METHOD_SET
-			  | GST_SEEK_FLAG_FLUSH,
-			  time * GST_SECOND);
-#elif HAVE_GSTREAMER_0_10
 	gst_element_seek (mp->priv->playbin, 1.0,
 			  GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH,
 			  GST_SEEK_TYPE_SET, time * GST_SECOND,
 			  GST_SEEK_TYPE_NONE, -1);
-#endif
 
 	if (mp->priv->playing) {
 		gst_element_set_state (mp->priv->playbin, GST_STATE_PLAYING);
@@ -1139,11 +1071,7 @@ rb_player_gst_get_time (RBPlayer *player)
 		gint64 position = -1;
 		GstFormat fmt = GST_FORMAT_TIME;
 
-#ifdef HAVE_GSTREAMER_0_8
-		gst_element_query (mp->priv->playbin, GST_QUERY_POSITION, &fmt, &position);
-#elif HAVE_GSTREAMER_0_10
 		gst_element_query_position (mp->priv->playbin, &fmt, &position);
-#endif
 		if (position != -1)
 			position /= GST_SECOND;
 
@@ -1151,3 +1079,304 @@ rb_player_gst_get_time (RBPlayer *player)
 	} else
 		return -1;
 }
+
+
+static gboolean
+rb_player_gst_add_tee (RBPlayerGstTee *player, GstElement *element)
+{
+	RBPlayerGst *mp;
+	GstElement *queue, *audioconvert, *bin;
+	GstPad *pad, *ghostpad;
+
+	mp = RB_PLAYER_GST (player);
+
+	if (mp->priv->tee == NULL) {
+		mp->priv->waiting_tees = g_list_prepend (mp->priv->waiting_tees, element);
+		return TRUE;
+	}
+
+	if (mp->priv->playing) {
+		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED) == GST_STATE_CHANGE_ASYNC) {
+			/* FIXME: Use a timeout on get_state. Post a GError somewhere on failed? */
+			if (gst_element_get_state (mp->priv->playbin, NULL, NULL, 3 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
+				g_warning ("Failed to pause pipeline before tee insertion");
+				return FALSE;
+			}
+		}
+	}
+
+	bin = gst_bin_new (NULL);
+	queue = gst_element_factory_make ("queue", NULL);
+	audioconvert = gst_element_factory_make ("audioconvert", NULL);
+
+	/* set up the element's containing bin */
+	gst_bin_add_many (GST_BIN (bin), queue, audioconvert, element, NULL);
+	gst_bin_add (GST_BIN (mp->priv->sinkbin), bin);
+	gst_element_link_many (queue, audioconvert, element, NULL);
+
+	/* link it to the tee */
+	pad = gst_element_get_pad (queue, "sink");
+	ghostpad = gst_ghost_pad_new ("sink", pad);
+	gst_element_add_pad (bin, ghostpad);
+	gst_object_unref (pad);
+
+	gst_element_link (mp->priv->tee, bin);
+
+	if (mp->priv->playing)
+		gst_element_set_state (mp->priv->playbin, GST_STATE_PLAYING);
+
+	return TRUE;
+}
+
+static gboolean
+rb_player_gst_remove_tee (RBPlayerGstTee *player, GstElement *element)
+{
+	RBPlayerGst *mp;
+	GstElement *bin;
+
+	mp = RB_PLAYER_GST (player);
+
+	if (mp->priv->tee == NULL) {
+		gst_object_sink (element);
+		mp->priv->waiting_tees = g_list_remove (mp->priv->waiting_tees, element);
+		return TRUE;
+	}
+
+	if (mp->priv->playing) {
+		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED) == GST_STATE_CHANGE_ASYNC) {
+			/* FIXME: Use a timeout on get_state. Post a GError somewhere on failed? */
+			if (gst_element_get_state (mp->priv->playbin, NULL, NULL, 3 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
+				g_warning ("Failed to pause pipeline before eee insertion");
+				return FALSE;
+			}
+		}
+	}
+
+	/* get the containing bin and unlink it */
+	bin = GST_ELEMENT (gst_element_get_parent (element));
+
+	if (gst_element_set_state (bin, GST_STATE_NULL) == GST_STATE_CHANGE_ASYNC) {
+		/* FIXME: Use a timeout on get_state. Post a GError somewhere on failed? */
+		if (gst_element_get_state (bin, NULL, NULL, 3 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
+			g_warning ("Failed to pause pipeline before tee insertion");
+			return FALSE;
+		}
+	}
+
+	gst_bin_remove (GST_BIN (mp->priv->sinkbin), bin);
+	gst_object_unref (bin);
+
+	if (mp->priv->playing)
+		gst_element_set_state (mp->priv->playbin, GST_STATE_PLAYING);
+
+	return TRUE;
+}
+
+static gboolean
+rb_player_gst_add_filter (RBPlayerGstFilter *player, GstElement *element)
+{
+	RBPlayerGst *mp;
+	GstElement *audioconvert, *bin;
+	GstPad *ghostpad, *realpad;
+	GstPad *binsinkpad, *binsrcpad;
+	gpointer element_sink_pad;
+	GstIterator *sink_pads;
+	gboolean sink_pad_found, stop_scan;
+	GstPadLinkReturn link;
+
+	mp = RB_PLAYER_GST (player);
+
+	if (mp->priv->filterbin == NULL) {
+		mp->priv->waiting_filters = g_list_prepend (mp->priv->waiting_filters, element);
+		return TRUE;
+	}
+
+	if (mp->priv->playing) {
+		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED) == GST_STATE_CHANGE_ASYNC) {
+			/* FIXME: Use a timeout on get_state. Post a GError somewhere on failed? */
+			if (gst_element_get_state (mp->priv->playbin, NULL, NULL, 3 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
+				g_warning ("Failed to pause pipeline before filter insertion");
+				return FALSE;
+			}
+		}
+	}
+
+	bin = gst_bin_new (NULL);
+	audioconvert = gst_element_factory_make ("audioconvert", NULL);
+
+	/* set up the element's containing bin */
+	rb_debug ("adding element %p and audioconvert to bin", element);
+	gst_bin_add_many (GST_BIN (bin), element, audioconvert, NULL);
+	gst_element_link_many (element, audioconvert, NULL);
+
+	/* ghost to the bin */
+	/* retrieve the first unliked source pad */
+	sink_pad_found = FALSE;
+	stop_scan = FALSE;
+	sink_pads = gst_element_iterate_sink_pads (element);
+	while (!sink_pad_found && !stop_scan) {
+		gpointer *esp_pointer = &element_sink_pad; /* stop type-punning warnings */
+		switch (gst_iterator_next (sink_pads, esp_pointer)) {
+			case GST_ITERATOR_OK:
+				sink_pad_found = !gst_pad_is_linked (GST_PAD(element_sink_pad));
+				break;
+			case GST_ITERATOR_RESYNC:
+				gst_iterator_resync (sink_pads);
+				break;
+			case GST_ITERATOR_ERROR:
+			case GST_ITERATOR_DONE:
+				stop_scan = TRUE;
+				break;
+		}
+	}
+	gst_iterator_free (sink_pads);
+
+	if (!sink_pad_found) {
+		g_warning ("Could not find a free sink pad on filter");
+		return FALSE;
+	}
+
+	binsinkpad = gst_ghost_pad_new ("sink", GST_PAD (element_sink_pad));
+	gst_element_add_pad (bin, binsinkpad);
+
+	realpad = gst_element_get_pad (audioconvert, "src");
+	binsrcpad = gst_ghost_pad_new ("src", realpad);
+	gst_element_add_pad (bin, binsrcpad);
+	gst_object_unref (realpad);
+
+	/* replace the filter chain ghost with the new bin */
+	gst_bin_add (GST_BIN (mp->priv->filterbin), bin);
+
+	ghostpad = gst_element_get_pad (mp->priv->filterbin, "src");
+	realpad = gst_ghost_pad_get_target (GST_GHOST_PAD (ghostpad));
+	gst_ghost_pad_set_target (GST_GHOST_PAD (ghostpad), binsrcpad);
+	gst_object_unref (ghostpad);
+
+	link = gst_pad_link (realpad, binsinkpad);
+	gst_object_unref (realpad);
+	if (link != GST_PAD_LINK_OK) {
+		g_warning ("could not link new filter into pipeline");
+		return FALSE;
+	}
+
+	if (mp->priv->playing)
+		gst_element_set_state (mp->priv->playbin, GST_STATE_PLAYING);
+
+	return TRUE;
+}
+
+static gboolean
+rb_player_gst_remove_filter (RBPlayerGstFilter *player, GstElement *element)
+{
+	RBPlayerGst *mp;
+	GstPad *mypad;
+	GstPad *prevpad, *nextpad;
+	GstPad *ghostpad;
+	GstPad *targetpad;
+	GstElement *bin;
+	gboolean result = TRUE;
+
+	mp = RB_PLAYER_GST (player);
+
+	if (mp->priv->filterbin == NULL) {
+		gst_object_sink (element);
+		mp->priv->waiting_filters = g_list_remove (mp->priv->waiting_filters, element);
+		return TRUE;
+	}
+
+	if (mp->priv->playing) {
+
+		/* it'd be more fun to do this by blocking a pad.. */
+
+		if (gst_element_set_state (mp->priv->playbin, GST_STATE_PAUSED) == GST_STATE_CHANGE_ASYNC) {
+			/* FIXME: Use a timeout on get_state. Post a GError somewhere on failed? */
+			if (gst_element_get_state (mp->priv->playbin, NULL, NULL, 3 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
+				g_warning ("Failed to pause pipeline before filter insertion");
+				return FALSE;
+			}
+		}
+	}
+
+	/* get the containing bin and unlink it */
+	bin = GST_ELEMENT (gst_element_get_parent (element));
+
+	if (gst_element_set_state (bin, GST_STATE_NULL) == GST_STATE_CHANGE_ASYNC) {
+		/* FIXME: Use a timeout on get_state. Post a GError somewhere on failed? */
+		if (gst_element_get_state (bin, NULL, NULL, 3 * GST_SECOND) != GST_STATE_CHANGE_SUCCESS) {
+			g_warning ("Failed to pause pipeline before filter insertion");
+			return FALSE;
+		}
+	}
+
+	mypad = gst_element_get_pad (bin, "sink");
+	prevpad = gst_pad_get_peer (mypad);
+	gst_pad_unlink (prevpad, mypad);
+	gst_object_unref (mypad);
+
+	ghostpad = gst_element_get_pad (bin, "src");
+	nextpad = gst_element_get_pad (mp->priv->filterbin, "src");
+
+	targetpad = gst_ghost_pad_get_target (GST_GHOST_PAD (nextpad));
+	if (targetpad == ghostpad) {
+		/* we are at the end of the filter chain, so redirect the ghostpad to the previous element */
+		gst_ghost_pad_set_target (GST_GHOST_PAD (nextpad), prevpad);
+	} else {
+		/* we are in the middle, so link the previous and next elements */
+		mypad = gst_element_get_pad (bin, "src");
+		gst_object_unref (nextpad);
+		nextpad = gst_pad_get_peer (mypad);
+		gst_pad_unlink (mypad, nextpad);
+		gst_object_unref (mypad);
+
+		if (gst_pad_link (prevpad, nextpad) != GST_PAD_LINK_OK)
+			result = FALSE;
+	}
+
+	gst_object_unref (nextpad);
+	gst_object_unref (prevpad);
+	gst_object_unref (ghostpad);
+	gst_object_unref (targetpad);
+
+	gst_bin_remove (GST_BIN (mp->priv->filterbin), bin);
+	gst_object_unref (bin);
+
+	if (mp->priv->playing)
+		gst_element_set_state (mp->priv->playbin, GST_STATE_PLAYING);
+
+	return result;
+}
+
+/*static gboolean
+rb_player_gst_add_data_tee (RBPlayerGstDataTee *player, GstElement *element)
+{
+
+}
+
+static gboolean
+rb_player_gst_remove_data_tee (RBPlayerGstDataTee *player, GstElement *element)
+{
+
+}*/
+
+static void
+rb_player_gst_filter_init (RBPlayerGstFilterIface *iface)
+{
+	iface->add_filter = rb_player_gst_add_filter;
+	iface->remove_filter = rb_player_gst_remove_filter;
+}
+
+static void
+rb_player_gst_tee_init (RBPlayerGstTeeIface *iface)
+{
+	iface->add_tee = rb_player_gst_add_tee;
+	iface->remove_tee = rb_player_gst_remove_tee;
+}
+
+/*static void
+rb_player_gst_data_tee_init (RBPlayerGstDataTeeIface *iface)
+{
+	iface->add_data_tee = rb_player_gst_add_data_tee;
+	iface->remove_data_tee = rb_player_gst_remove_data_tee;
+}*/
+
+
