@@ -1,5 +1,5 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
- * 
+ *
  *  arch-tag: Implementation of preferences dialog object
  *
  *  Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org>
@@ -60,7 +60,17 @@ void rb_shell_preferences_browser_views_activated_cb (GtkWidget *widget,
 						      RBShellPreferences *shell_preferences);
 static void rb_shell_preferences_toolbar_style_cb (GtkComboBox *box,
 						   RBShellPreferences *preferences);
-
+#ifdef HAVE_GSTREAMER_0_10_XFADE
+static void rb_shell_preferences_player_backend_cb (GtkToggleButton *button,
+						    RBShellPreferences *preferences);
+static void rb_shell_preferences_album_crossfade_cb (GtkToggleButton *button,
+						     RBShellPreferences *preferences);
+static void rb_shell_preferences_transition_duration_cb (GtkRange *range,
+							 RBShellPreferences *preferences);
+static void rb_shell_preferences_network_buffer_size_cb (GtkRange *range,
+							 RBShellPreferences *preferences);
+static void update_playback_prefs_sensitivity (RBShellPreferences *preferences);
+#endif
 
 enum
 {
@@ -86,6 +96,13 @@ struct RBShellPreferencesPrivate
 	GtkWidget *quality_check;
 	GtkWidget *year_check;
 
+#ifdef HAVE_GSTREAMER_0_10_XFADE
+	GtkWidget *xfade_backend_check;
+	GtkWidget *album_crossfade_check;
+	GtkWidget *transition_duration;
+	GtkWidget *network_buffer_size;
+#endif
+
 	GSList *browser_views_group;
 
 	GtkWidget *toolbar_style_menu;
@@ -132,11 +149,11 @@ rb_shell_preferences_init (RBShellPreferences *shell_preferences)
 
 	shell_preferences->priv = RB_SHELL_PREFERENCES_GET_PRIVATE (shell_preferences);
 
-	g_signal_connect_object (G_OBJECT (shell_preferences),
+	g_signal_connect_object (shell_preferences,
 				 "delete_event",
 				 G_CALLBACK (rb_shell_preferences_window_delete_cb),
 				 shell_preferences, 0);
-	g_signal_connect_object (G_OBJECT (shell_preferences),
+	g_signal_connect_object (shell_preferences,
 				 "response",
 				 G_CALLBACK (rb_shell_preferences_response_cb),
 				 shell_preferences, 0);
@@ -147,7 +164,7 @@ rb_shell_preferences_init (RBShellPreferences *shell_preferences)
 	tmp = gtk_dialog_add_button (GTK_DIALOG (shell_preferences),
 			              GTK_STOCK_HELP,
 			              GTK_RESPONSE_HELP);
-	g_signal_connect_object (G_OBJECT (tmp), "clicked",
+	g_signal_connect_object (tmp, "clicked",
 				 G_CALLBACK (help_cb), shell_preferences, 0);
 	gtk_dialog_set_default_response (GTK_DIALOG (shell_preferences),
 					 GTK_RESPONSE_CLOSE);
@@ -170,7 +187,7 @@ rb_shell_preferences_init (RBShellPreferences *shell_preferences)
 				shell_preferences);
 
 	rb_glade_boldify_label (xml, "visible_columns_label");
-	
+
 	/* Columns */
 	shell_preferences->priv->artist_check =
 		glade_xml_get_widget (xml, "artist_check");
@@ -200,7 +217,7 @@ rb_shell_preferences_init (RBShellPreferences *shell_preferences)
 
 	tmp = glade_xml_get_widget (xml, "library_browser_views_radio");
 	shell_preferences->priv->browser_views_group =
-		g_slist_reverse (g_slist_copy (gtk_radio_button_get_group 
+		g_slist_reverse (g_slist_copy (gtk_radio_button_get_group
 					       (GTK_RADIO_BUTTON (tmp))));
 
 	gtk_notebook_append_page (GTK_NOTEBOOK (shell_preferences->priv->notebook),
@@ -217,11 +234,62 @@ rb_shell_preferences_init (RBShellPreferences *shell_preferences)
 	g_signal_connect_object (G_OBJECT (shell_preferences->priv->toolbar_style_menu),
 				 "changed", G_CALLBACK (rb_shell_preferences_toolbar_style_cb),
 				 shell_preferences, 0);
-	
+
 	eel_gconf_notification_add (CONF_UI_DIR,
 				    (GConfClientNotifyFunc) rb_shell_preferences_ui_pref_changed,
 				    shell_preferences);
-	g_object_unref (G_OBJECT (xml));
+	g_object_unref (xml);
+
+#ifdef HAVE_GSTREAMER_0_10_XFADE
+	/* playback preferences */
+	xml = rb_glade_xml_new ("playback-prefs.glade",
+				"playback_prefs_box",
+				shell_preferences);
+
+	rb_glade_boldify_label (xml, "backend_label");
+	rb_glade_boldify_label (xml, "crossfade_type_label");
+	rb_glade_boldify_label (xml, "duration_label");
+	rb_glade_boldify_label (xml, "buffer_size_label");
+
+	shell_preferences->priv->xfade_backend_check =
+		glade_xml_get_widget (xml, "use_xfade_backend");
+	shell_preferences->priv->album_crossfade_check =
+		glade_xml_get_widget (xml, "album_check");
+	shell_preferences->priv->transition_duration =
+		glade_xml_get_widget (xml, "duration");
+	shell_preferences->priv->network_buffer_size =
+		glade_xml_get_widget (xml, "network_buffer_size");
+
+	g_signal_connect_object (shell_preferences->priv->xfade_backend_check,
+				 "toggled",
+				 G_CALLBACK (rb_shell_preferences_player_backend_cb),
+				 shell_preferences, 0);
+
+	g_signal_connect_object (shell_preferences->priv->album_crossfade_check,
+				 "toggled",
+				 G_CALLBACK (rb_shell_preferences_album_crossfade_cb),
+				 shell_preferences, 0);
+
+	g_signal_connect_object (shell_preferences->priv->transition_duration,
+				 "value-changed",
+				 G_CALLBACK (rb_shell_preferences_transition_duration_cb),
+				 shell_preferences, 0);
+
+	g_signal_connect_object (shell_preferences->priv->network_buffer_size,
+				 "value-changed",
+				 G_CALLBACK (rb_shell_preferences_network_buffer_size_cb),
+				 shell_preferences, 0);
+
+	gtk_notebook_append_page (GTK_NOTEBOOK (shell_preferences->priv->notebook),
+				  glade_xml_get_widget (xml, "playback_prefs_box"),
+				  gtk_label_new (_("Playback")));
+	g_object_unref (xml);
+#endif
+
+	eel_gconf_notification_add (CONF_PLAYER_DIR,
+				    (GConfClientNotifyFunc) rb_shell_preferences_ui_pref_changed,
+				    shell_preferences);
+
 	rb_shell_preferences_sync (shell_preferences);
 }
 
@@ -246,7 +314,7 @@ rb_shell_preferences_append_page (RBShellPreferences *prefs,
 				  GtkWidget *widget)
 {
 	GtkWidget *label;
-		
+
 	label = gtk_label_new (name);
 	gtk_notebook_append_page (GTK_NOTEBOOK (prefs->priv->notebook),
 				  widget,
@@ -270,158 +338,6 @@ rb_shell_preferences_append_view_page (RBShellPreferences *prefs,
 	rb_shell_preferences_append_page (prefs, name, widget);
 }
 
-#ifdef WITH_DAAP_SUPPORT
-
-static void
-share_check_button_toggled_cb (GtkToggleButton *button,
-			       GtkWidget *widget)
-{
-	gboolean b;
-
-	b = gtk_toggle_button_get_active (button);
-
-	eel_gconf_set_boolean (CONF_DAAP_ENABLE_SHARING, b);
-
-	gtk_widget_set_sensitive (widget, b);
-	
-	return;
-}
-
-static void
-password_check_button_toggled_cb (GtkToggleButton *button,
-				  GtkWidget *widget)
-{
-	gboolean b;
-
-	b = gtk_toggle_button_get_active (button);
-
-	eel_gconf_set_boolean (CONF_DAAP_REQUIRE_PASSWORD, b);
-
-	gtk_widget_set_sensitive (widget, b);
-	
-	return;
-}
-
-static gboolean
-share_name_entry_focus_out_event_cb (GtkEntry *entry,
-				     GdkEventFocus *event,
-				     gpointer data)
-{
-	gboolean    changed;
-	const char *name;
-	char       *old_name;
-
-	name = gtk_entry_get_text (entry);
-	old_name = eel_gconf_get_string (CONF_DAAP_SHARE_NAME);
-
-	if (name == NULL && old_name == NULL) {
-		changed = FALSE;
-	} else if (name == NULL || old_name == NULL) {
-		changed = TRUE;
-	} else if (strcmp (name, old_name) != 0) {
-		changed = TRUE;
-	} else {
-		changed = FALSE;
-	}
-
-	if (changed) {
-		eel_gconf_set_string (CONF_DAAP_SHARE_NAME, name);
-	}
-
-	g_free (old_name);
-
-	return FALSE;
-}
-
-static gboolean
-share_password_entry_focus_out_event_cb (GtkEntry *entry,
-					 GdkEventFocus *event,
-					 gpointer data)
-{
-	gboolean    changed;
-	const char *pw;
-	char       *old_pw;
-
-	pw = gtk_entry_get_text (entry);
-	old_pw = eel_gconf_get_string (CONF_DAAP_SHARE_PASSWORD);
-
-	if (pw == NULL && old_pw == NULL) {
-		changed = FALSE;
-	} else if (pw == NULL || old_pw == NULL) {
-		changed = TRUE;
-	} else if (strcmp (pw, old_pw) != 0) {
-		changed = TRUE;
-	} else {
-		changed = FALSE;
-	}
-
-	if (changed) {
-		eel_gconf_set_string (CONF_DAAP_SHARE_PASSWORD, pw);
-	}
-
-	g_free (old_pw);
-
-	return FALSE;
-}
-
-static void
-add_daap_preferences (RBShellPreferences *shell_preferences)
-{
-	GladeXML *xml;
-	GtkWidget *check;
-	GtkWidget *name_entry;
-	GtkWidget *password_entry;
-	GtkWidget *password_check;
-	GtkWidget *box;
-	gboolean sharing_enabled;
-	gboolean require_password;
-
-	char *name;
-	char *password;
-
-	xml = rb_glade_xml_new ("daap-prefs.glade",
-				"daap_vbox",
-				shell_preferences);
-
-	check = glade_xml_get_widget (xml, "daap_enable_check");
-	password_check = glade_xml_get_widget (xml, "daap_password_check");
-	name_entry = glade_xml_get_widget (xml, "daap_name_entry");
-	password_entry = glade_xml_get_widget (xml, "daap_password_entry");
-	box = glade_xml_get_widget (xml, "daap_box");
-	
-	sharing_enabled = eel_gconf_get_boolean (CONF_DAAP_ENABLE_SHARING);
-	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (check), sharing_enabled);
-	g_signal_connect (check, "toggled", G_CALLBACK (share_check_button_toggled_cb), box);
-
-	require_password = eel_gconf_get_boolean (CONF_DAAP_REQUIRE_PASSWORD);
-	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (password_check), require_password);
-	g_signal_connect (password_check, "toggled", G_CALLBACK (password_check_button_toggled_cb), password_entry);
-	
-	name = eel_gconf_get_string (CONF_DAAP_SHARE_NAME);
-	if (name != NULL) {
-		gtk_entry_set_text (GTK_ENTRY (name_entry), name);
-	}
-	g_free (name);
-	g_signal_connect (name_entry, "focus-out-event",
-			  G_CALLBACK (share_name_entry_focus_out_event_cb), NULL);	
-
-	password = eel_gconf_get_string (CONF_DAAP_SHARE_PASSWORD);
-	if (password != NULL) {
-		gtk_entry_set_text (GTK_ENTRY (password_entry), password);
-	}
-	g_free (password);
-	g_signal_connect (password_entry, "focus-out-event",
-			  G_CALLBACK (share_password_entry_focus_out_event_cb), NULL);	
-
-	gtk_widget_set_sensitive (box, sharing_enabled);
-	gtk_widget_set_sensitive (password_entry, require_password);
-
-	gtk_notebook_append_page (GTK_NOTEBOOK (shell_preferences->priv->notebook),
-				  glade_xml_get_widget (xml, "daap_vbox"),
-				  gtk_label_new (_("Sharing")));
-}
-#endif /* WITH_DAAP_SUPPORT */
-
 GtkWidget *
 rb_shell_preferences_new (GList *views)
 {
@@ -434,18 +350,15 @@ rb_shell_preferences_new (GList *views)
 
 	for (; views; views = views->next)
 	{
-		const char *name = NULL;
-		g_object_get (G_OBJECT (views->data), "name", &name, NULL);
+		char *name = NULL;
+		g_object_get (views->data, "name", &name, NULL);
 		g_assert (name != NULL);
 		rb_shell_preferences_append_view_page (shell_preferences,
 						       name,
 						       RB_SOURCE (views->data));
+		g_free (name);
 	}
 
-#ifdef WITH_DAAP_SUPPORT
-	add_daap_preferences (shell_preferences);
-#endif /* WITH_DAAP_SUPPORT */
-
 	return GTK_WIDGET (shell_preferences);
 }
 
@@ -516,7 +429,7 @@ rb_shell_preferences_column_check_changed_cb (GtkCheckButton *butt,
 		g_assert_not_reached ();
 
 	rb_debug ("\"%s\" changed, current cols are \"%s\"", colname, currentcols);
-	
+
 	/* Append this if we want it */
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (butt))) {
 		g_string_append (newcolumns, colname);
@@ -528,7 +441,7 @@ rb_shell_preferences_column_check_changed_cb (GtkCheckButton *butt,
 		if (strcmp (colnames[i], colname)) {
 			g_string_append (newcolumns, colnames[i]);
 			if (colnames[i+1] != NULL)
-				g_string_append (newcolumns, ",");				
+				g_string_append (newcolumns, ",");
 		}
 	}
 
@@ -545,8 +458,8 @@ rb_shell_preferences_sync_column_button (RBShellPreferences *preferences,
 	g_signal_handlers_block_by_func (G_OBJECT (button),
 					 rb_shell_preferences_column_check_changed_cb,
 					 preferences);
-	
-	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), 
+
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
 				      strstr (columns, propid) != NULL);
 
 	g_signal_handlers_unblock_by_func (G_OBJECT (button),
@@ -564,7 +477,7 @@ rb_shell_preferences_sync (RBShellPreferences *shell_preferences)
 	shell_preferences->priv->loading = TRUE;
 
 	rb_debug ("syncing prefs");
-	
+
 	columns = eel_gconf_get_string (CONF_UI_COLUMNS_SETUP);
 	if (columns != NULL)
 	{
@@ -624,6 +537,29 @@ rb_shell_preferences_sync (RBShellPreferences *shell_preferences)
 					   G_CALLBACK (rb_shell_preferences_toolbar_style_cb),
 					   shell_preferences);
 
+#ifdef HAVE_GSTREAMER_0_10_XFADE
+	{
+		gboolean b;
+		float duration;
+		int buffer_size;
+
+		/* player preferences */
+		b = eel_gconf_get_boolean (CONF_PLAYER_USE_XFADE_BACKEND);
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (shell_preferences->priv->xfade_backend_check), b);
+
+		b = eel_gconf_get_boolean (CONF_PLAYER_TRANSITION_ALBUM_CHECK);
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (shell_preferences->priv->album_crossfade_check), !b);
+
+		duration = eel_gconf_get_float (CONF_PLAYER_TRANSITION_TIME);
+		gtk_range_set_value (GTK_RANGE (shell_preferences->priv->transition_duration), duration);
+
+		buffer_size = eel_gconf_get_integer (CONF_PLAYER_NETWORK_BUFFER_SIZE);
+		gtk_range_set_value (GTK_RANGE (shell_preferences->priv->network_buffer_size), buffer_size);
+
+		update_playback_prefs_sensitivity (shell_preferences);
+	}
+#endif
+
 	shell_preferences->priv->loading = FALSE;
 }
 
@@ -655,3 +591,59 @@ rb_shell_preferences_browser_views_activated_cb (GtkWidget *widget,
 	eel_gconf_set_integer (CONF_UI_BROWSER_VIEWS, index);
 }
 
+#ifdef HAVE_GSTREAMER_0_10_XFADE
+
+static void
+update_playback_prefs_sensitivity (RBShellPreferences *preferences)
+{
+	gboolean backend;
+
+	backend = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (preferences->priv->xfade_backend_check));
+
+	gtk_widget_set_sensitive (preferences->priv->album_crossfade_check, backend);
+	gtk_widget_set_sensitive (preferences->priv->network_buffer_size, backend);
+	gtk_widget_set_sensitive (preferences->priv->transition_duration, backend);
+}
+
+static void
+rb_shell_preferences_player_backend_cb (GtkToggleButton *button,
+					RBShellPreferences *preferences)
+{
+	update_playback_prefs_sensitivity (preferences);
+
+	eel_gconf_set_boolean (CONF_PLAYER_USE_XFADE_BACKEND,
+			       gtk_toggle_button_get_active (button));
+}
+
+static void
+rb_shell_preferences_album_crossfade_cb (GtkToggleButton *button,
+					 RBShellPreferences *preferences)
+{
+	/* sense is inverted here because the wording sounded more natural
+	 * this way.  "[ ] Don't do x" sucks.
+	 */
+	eel_gconf_set_boolean (CONF_PLAYER_TRANSITION_ALBUM_CHECK,
+			       !gtk_toggle_button_get_active (button));
+}
+
+static void
+rb_shell_preferences_transition_duration_cb (GtkRange *range,
+					     RBShellPreferences *preferences)
+{
+	gdouble v;
+
+	v = gtk_range_get_value (range);
+	eel_gconf_set_float (CONF_PLAYER_TRANSITION_TIME, (float)v);
+}
+
+static void
+rb_shell_preferences_network_buffer_size_cb (GtkRange *range,
+					     RBShellPreferences *preferences)
+{
+	gdouble v;
+
+	v = gtk_range_get_value (range);
+	eel_gconf_set_integer (CONF_PLAYER_NETWORK_BUFFER_SIZE, (int)v);
+}
+
+#endif
