@@ -37,7 +37,7 @@
 
 struct RBSourceListModelPrivate
 {
-	GtkTreeRowReference *groups[RB_SOURCELIST_GROUP_LAST];
+	gpointer dummy;
 };
 
 #define RB_SOURCELIST_MODEL_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), RB_TYPE_SOURCELIST_MODEL, RBSourceListModelPrivate))
@@ -192,15 +192,10 @@ static void
 rb_sourcelist_model_finalize (GObject *object)
 {
 	RBSourceListModel *model;
-	int i;
 
 	g_return_if_fail (RB_IS_SOURCELIST_MODEL (object));
 	model = RB_SOURCELIST_MODEL (object);
 
-	for (i = 0; i < RB_SOURCELIST_GROUP_LAST; i++) {
-		gtk_tree_row_reference_free (model->priv->groups[i]);
-	}
-
 	G_OBJECT_CLASS (rb_sourcelist_model_parent_class)->finalize (object);
 }
 
@@ -209,7 +204,6 @@ rb_sourcelist_model_new (void)
 {
 	RBSourceListModel *model;
 	GtkTreeStore *store;
-	int i;
  	GType *column_types = g_new (GType, RB_SOURCELIST_MODEL_N_COLUMNS);
 
 	column_types[RB_SOURCELIST_MODEL_COLUMN_PLAYING] = G_TYPE_BOOLEAN;
@@ -218,6 +212,8 @@ rb_sourcelist_model_new (void)
 	column_types[RB_SOURCELIST_MODEL_COLUMN_SOURCE] = G_TYPE_OBJECT;
 	column_types[RB_SOURCELIST_MODEL_COLUMN_ATTRIBUTES] = PANGO_TYPE_ATTR_LIST;
 	column_types[RB_SOURCELIST_MODEL_COLUMN_VISIBILITY] = G_TYPE_BOOLEAN;
+	column_types[RB_SOURCELIST_MODEL_COLUMN_IS_GROUP] = G_TYPE_BOOLEAN;
+	column_types[RB_SOURCELIST_MODEL_COLUMN_GROUP_CATEGORY] = RB_TYPE_SOURCE_GROUP_CATEGORY;
 	store = gtk_tree_store_newv (RB_SOURCELIST_MODEL_N_COLUMNS,
 				     column_types);
 
@@ -227,23 +223,6 @@ rb_sourcelist_model_new (void)
 						   NULL));
 	g_object_unref (store);
 
-	/* create marker rows used to separate source groups */
-	for (i = 0; i < RB_SOURCELIST_GROUP_LAST; i++) {
-		GtkTreeIter iter;
-		GtkTreePath *path;
-
-		gtk_tree_store_append (store, &iter, NULL);
-		gtk_tree_store_set (store, &iter,
-				    RB_SOURCELIST_MODEL_COLUMN_NAME, "",
-				    RB_SOURCELIST_MODEL_COLUMN_SOURCE, NULL,
-				    RB_SOURCELIST_MODEL_COLUMN_VISIBILITY, FALSE,
-				    -1);
-
-		path = gtk_tree_model_get_path (GTK_TREE_MODEL (store), &iter);
-		model->priv->groups[i] = gtk_tree_row_reference_new (GTK_TREE_MODEL (store), path);
-		gtk_tree_path_free (path);
-	}
-
 	/* ensure the group markers get updated as sources are added and removed */
 	g_signal_connect_object (G_OBJECT (store), "row-inserted",
 				 G_CALLBACK (rb_sourcelist_model_row_inserted_cb),
@@ -261,57 +240,18 @@ rb_sourcelist_model_new (void)
 	return GTK_TREE_MODEL (model);
 }
 
-static gboolean
-real_row_is_separator (RBSourceListModel *model, GtkTreeIter *iter)
-{
-	int i;
-	GtkTreePath *path;
-	GtkTreeModel *real_model;
-	GtkTreePath *group_path;
-
-	real_model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (model));
-	path = gtk_tree_model_get_path (real_model, iter);
-
-	/* -1 here because the last group marker is always the last row in
-	 * the model, and therefore can never be shown.
-	 */
-	for (i = 0; i < RB_SOURCELIST_GROUP_LAST-1; i++) {
-		group_path = rb_sourcelist_model_get_group_path (model, i);
-
-		if (gtk_tree_path_compare (path, group_path) == 0) {
-			/* okay, we know this is a group marker.
-			 * if the next row is not a group marker,
-			 * then we should show this row as a separator.
-			 */
-			gboolean separator = FALSE;
-
-			gtk_tree_path_next (path);
-
-			gtk_tree_path_free (group_path);
-			group_path = rb_sourcelist_model_get_group_path (model, i+1);
-			separator = (gtk_tree_path_compare (path, group_path) != 0);
-
-			gtk_tree_path_free (group_path);
-			gtk_tree_path_free (path);
-			return separator;
-		}
-		gtk_tree_path_free (group_path);
-	}
-
-	gtk_tree_path_free (path);
-
-	return FALSE;
-}
-
 static gboolean
 rb_sourcelist_model_is_row_visible (GtkTreeModel *model,
 				    GtkTreeIter *iter,
 				    RBSourceListModel *sourcelist)
 {
 	RBSource *source;
+	gboolean  visibility;
 
 	gtk_tree_model_get (GTK_TREE_MODEL (model), iter,
-			    RB_SOURCELIST_MODEL_COLUMN_SOURCE, &source, -1);
+			    RB_SOURCELIST_MODEL_COLUMN_SOURCE, &source,
+			    RB_SOURCELIST_MODEL_COLUMN_VISIBILITY, &visibility,
+			    -1);
 
 	if (source != NULL) {
 		gboolean visible;
@@ -321,51 +261,10 @@ rb_sourcelist_model_is_row_visible (GtkTreeModel *model,
 
 		return visible;
 	} else {
-		return real_row_is_separator (sourcelist, iter);
-	}
-}
-
-gboolean
-rb_sourcelist_model_row_is_separator (GtkTreeModel *model,
-				      GtkTreeIter *iter,
-				      RBSourceListModel *sourcelist)
-{
-	GtkTreeIter real_iter;
-	RBSource *source;
-
-	/* rows with actual sources are never separators */
-	gtk_tree_model_get (GTK_TREE_MODEL (model), iter,
-			    RB_SOURCELIST_MODEL_COLUMN_SOURCE, &source, -1);
-	if (source != NULL) {
-		g_object_unref (source);
-		return FALSE;
+		return visibility;
 	}
 
-	gtk_tree_model_filter_convert_iter_to_child_iter (GTK_TREE_MODEL_FILTER (sourcelist),
-							  &real_iter,
-							  iter);
-	return real_row_is_separator (sourcelist, &real_iter);
-}
-
-static int
-get_group_for_path (RBSourceListModel *model, GtkTreePath *path)
-{
-	GtkTreePath *group_path;
-	gboolean found = FALSE;
-	int i;
-
-	for (i=0; i < RB_SOURCELIST_GROUP_LAST; i++) {
-		group_path = rb_sourcelist_model_get_group_path (model, i);
-		g_assert (group_path);
-
-		found = (gtk_tree_path_compare (group_path, path) == 1);
-		gtk_tree_path_free (group_path);
-
-		if (found)
-			return i;
-	}
-
-	g_assert_not_reached ();
+	return TRUE;
 }
 
 static gboolean
@@ -411,69 +310,8 @@ rb_sourcelist_model_drag_data_received (RbTreeDragDest *drag_dest,
         }
 
 	if (selection_data->type == gdk_atom_intern ("application/x-rhythmbox-source", TRUE)) {
-		GtkTreePath *path;
-		GtkTreePath *real_dest;
-		char *path_str;
-		GtkTreeIter iter, real_iter;
-		GtkTreeIter real_dest_iter;
-		GtkTreeModel *real_model;
-		RBSource *source;
-		RBSourceListGroup group;
-		int dest_group;
-
-		if (!dest)
-			return FALSE;
-
-		real_model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (model));
-
-		path_str = g_strndup ((char *) selection_data->data, selection_data->length);
-
-		path = gtk_tree_path_new_from_string (path_str);
-		gtk_tree_model_get_iter (GTK_TREE_MODEL (model),
-					 &iter, path);
-		gtk_tree_model_filter_convert_iter_to_child_iter (GTK_TREE_MODEL_FILTER (model),
-								  &real_iter, &iter);
-
-		real_dest =
-			gtk_tree_model_filter_convert_path_to_child_path (GTK_TREE_MODEL_FILTER (model),
-									  dest);
-
-		gtk_tree_model_get (GTK_TREE_MODEL (real_model), &real_iter,
-				    RB_SOURCELIST_MODEL_COLUMN_SOURCE, &source,
-				    -1);
-		g_object_get (source, "sourcelist-group", &group, NULL);
-		g_object_unref (source);
-
-		/* restrict sources to within their group */
-		dest_group = get_group_for_path (model, real_dest);
-		if (dest_group < group) {
-			gtk_tree_path_free (real_dest);
-			real_dest = rb_sourcelist_model_get_group_path (model, group-1);
-			pos = GTK_TREE_VIEW_DROP_AFTER;
-		} else if (dest_group > group) {
-			gtk_tree_path_free (real_dest);
-			real_dest = rb_sourcelist_model_get_group_path (model, group);
-			pos = GTK_TREE_VIEW_DROP_BEFORE;
-		}
-
-		if (gtk_tree_model_get_iter (GTK_TREE_MODEL (real_model),
-					     &real_dest_iter, real_dest)) {
-
-			if (pos == GTK_TREE_VIEW_DROP_AFTER)
-				gtk_tree_store_move_after (GTK_TREE_STORE (real_model),
-							   &real_iter, &real_dest_iter);
-			else
-				gtk_tree_store_move_before (GTK_TREE_STORE (real_model),
-							    &real_iter, &real_dest_iter);
-
-		} else {
-			gtk_tree_store_move_before (GTK_TREE_STORE (real_model),
-						    &real_iter, NULL);
-		}
-
-		gtk_tree_path_free (real_dest);
-		gtk_tree_path_free (path);
-		g_free (path_str);
+		/* don't support dnd of sources */
+		return FALSE;
 	}
 
 	return FALSE;
@@ -489,6 +327,7 @@ rb_sourcelist_model_row_drop_possible (RbTreeDragDest *drag_dest,
 
 	rb_debug ("row drop possible");
 	g_return_val_if_fail (RB_IS_SOURCELIST_MODEL (drag_dest), FALSE);
+
 	model = RB_SOURCELIST_MODEL (drag_dest);
 
 	if (!dest)
@@ -523,33 +362,6 @@ path_is_droppable (RBSourceListModel *model,
 	return res;
 }
 
-static gboolean
-path_is_reorderable (RBSourceListModel *model,
-		     GtkTreePath *dest)
-{
-	GtkTreeIter iter;
-
-	if (gtk_tree_model_get_iter (GTK_TREE_MODEL (model), &iter, dest)) {
-		RBSource *source;
-		RBSourceListGroup group;
-
-		gtk_tree_model_get (GTK_TREE_MODEL (model), &iter,
-				    RB_SOURCELIST_MODEL_COLUMN_SOURCE, &source, -1);
-
-		if (source == NULL)
-			return FALSE;
-
-		g_object_get (source, "sourcelist-group", &group, NULL);
-		g_object_unref (source);
-
-		/* fixed and transient sources are not reorderable, everything else is */
-		return (group != RB_SOURCELIST_GROUP_FIXED &&
-			group != RB_SOURCELIST_GROUP_TRANSIENT);
-	}
-
-	return FALSE;
-}
-
 static gboolean
 rb_sourcelist_model_row_drop_position (RbTreeDragDest   *drag_dest,
 				       GtkTreePath       *dest_path,
@@ -560,25 +372,7 @@ rb_sourcelist_model_row_drop_position (RbTreeDragDest   *drag_dest,
 
 	if (g_list_find (targets, gdk_atom_intern ("application/x-rhythmbox-source", TRUE)) && dest_path) {
 		rb_debug ("application/x-rhythmbox-source type");
-		if (!path_is_reorderable (RB_SOURCELIST_MODEL (model), dest_path)) {
-			GtkTreePath *test_path = gtk_tree_path_copy (dest_path);
-			gtk_tree_path_next (test_path);
-			gboolean ret = FALSE;
-			if (path_is_reorderable (RB_SOURCELIST_MODEL (model), test_path)) {
-				*pos = GTK_TREE_VIEW_DROP_AFTER;
-				ret = TRUE;
-			}
-
-			gtk_tree_path_free (test_path);
-			return ret;
-		}
-
-		if (*pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
-			*pos = GTK_TREE_VIEW_DROP_BEFORE;
-		else if (*pos == GTK_TREE_VIEW_DROP_INTO_OR_AFTER)
-			*pos = GTK_TREE_VIEW_DROP_AFTER;
-
-		return TRUE;
+		return FALSE;
 	}
 
 	if (g_list_find (targets, gdk_atom_intern ("text/uri-list", TRUE)) ||
@@ -634,29 +428,6 @@ static gboolean
 rb_sourcelist_model_row_draggable (RbTreeDragSource *drag_source,
 				   GList *path_list)
 {
-	GtkTreeIter iter;
-	GtkTreePath *path;
-	GtkTreeModel *model = GTK_TREE_MODEL (drag_source);
-
-	/* we don't support multi selection */
-	g_return_val_if_fail (g_list_length (path_list) == 1, FALSE);
-
-	path = gtk_tree_row_reference_get_path (path_list->data);
-
-	if (path && gtk_tree_model_get_iter (model, &iter, path)) {
-		RBSource *source;
-		RBSourceListGroup group;
-
-		gtk_tree_model_get (GTK_TREE_MODEL (model), &iter,
-				    RB_SOURCELIST_MODEL_COLUMN_SOURCE, &source, -1);
-
-		g_object_get (source, "sourcelist-group", &group, NULL);
-		g_object_unref (source);
-
-		return (group != RB_SOURCELIST_GROUP_FIXED &&
-			group != RB_SOURCELIST_GROUP_TRANSIENT);
-	}
-
 	return FALSE;
 }
 
@@ -775,9 +546,28 @@ rb_sourcelist_model_row_deleted_cb (GtkTreeModel *model,
 	gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (sourcelist));
 }
 
-GtkTreePath *
-rb_sourcelist_model_get_group_path (RBSourceListModel *sourcelist,
-				    RBSourceListGroup group)
+/* This should really be standard. */
+#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }
+
+GType
+rb_sourcelist_model_column_get_type (void)
 {
-	return gtk_tree_row_reference_get_path (sourcelist->priv->groups[group]);
+	static GType etype = 0;
+
+	if (etype == 0)	{
+		static const GEnumValue values[] = {
+			ENUM_ENTRY (RB_SOURCELIST_MODEL_COLUMN_PLAYING, "Playing"),
+			ENUM_ENTRY (RB_SOURCELIST_MODEL_COLUMN_PIXBUF, "Pixbuf Icon"),
+			ENUM_ENTRY (RB_SOURCELIST_MODEL_COLUMN_NAME, "Name"),
+			ENUM_ENTRY (RB_SOURCELIST_MODEL_COLUMN_SOURCE, "Source"),
+			ENUM_ENTRY (RB_SOURCELIST_MODEL_COLUMN_ATTRIBUTES, "Attributes"),
+			ENUM_ENTRY (RB_SOURCELIST_MODEL_COLUMN_VISIBILITY, "Visibility"),
+			{ 0, 0, 0 }
+		};
+
+		etype = g_enum_register_static ("RBSourceListModelColumn", values);
+	}
+
+	return etype;
 }
+
